-----------------------------------------------------------------------------
-- |
-- Module      :  Graphics.Rendering.Chart.Plot
-- Copyright   :  (c) Tim Docker 2006
-- License     :  BSD-style (see chart/COPYRIGHT)
--
-- Code to calculate and render various types of plots.
--

module Graphics.Rendering.Chart.Plot(
    module Graphics.Rendering.Chart.Plot.Types,
    module Graphics.Rendering.Chart.Plot.Lines,
    module Graphics.Rendering.Chart.Plot.Vectors,
    module Graphics.Rendering.Chart.Plot.Points,
    module Graphics.Rendering.Chart.Plot.FillBetween,
    module Graphics.Rendering.Chart.Plot.ErrBars,
    module Graphics.Rendering.Chart.Plot.Candle,
    module Graphics.Rendering.Chart.Plot.Bars,
    module Graphics.Rendering.Chart.Plot.Hidden,
    module Graphics.Rendering.Chart.Plot.Annotation,
    module Graphics.Rendering.Chart.Plot.AreaSpots,
    module Graphics.Rendering.Chart.Plot.Pie,
    module Graphics.Rendering.Chart.Plot.Histogram,
) where

import Graphics.Rendering.Chart.Plot.Types
import Graphics.Rendering.Chart.Plot.Lines
import Graphics.Rendering.Chart.Plot.Vectors
import Graphics.Rendering.Chart.Plot.Points
import Graphics.Rendering.Chart.Plot.FillBetween
import Graphics.Rendering.Chart.Plot.ErrBars
import Graphics.Rendering.Chart.Plot.Candle
import Graphics.Rendering.Chart.Plot.Bars
import Graphics.Rendering.Chart.Plot.Hidden
import Graphics.Rendering.Chart.Plot.Annotation
import Graphics.Rendering.Chart.Plot.AreaSpots
import Graphics.Rendering.Chart.Plot.Pie
import Graphics.Rendering.Chart.Plot.Histogram
